from functools import singledispatch

from ieml.exceptions import TermNotFoundInDictionary
from .version import DictionaryVersion
from .script import Script
from .terms import Term
from .dictionary import Dictionary


def term(arg, dictionary=None):
    if not isinstance(dictionary, Dictionary):
        if isinstance(dictionary, (str, DictionaryVersion)):
            dictionary = Dictionary(dictionary)
        else:
            dictionary = Dictionary()

    try:
        return _term(arg, dictionary)
    except KeyError:
        raise TermNotFoundInDictionary(arg, dictionary)

@singledispatch
def _term(arg, dictionary):
    raise NotImplemented

_term.register(Term, lambda arg, dictionary: arg)
_term.register(int, lambda arg, dictionary: dictionary.index[arg])
_term.register(Script, lambda arg, dictionary: dictionary.terms[arg])


@_term.register(str)
def _term_str(arg, dictionary=None):
    if arg[0] == '[' and arg[-1] == ']':
        arg = arg[1:-1]
    return dictionary.terms[arg]




import ply.lex as lxr
import logging

tokens = (
   'IEML_OBJECT',
   # 'L_CURLY_BRACKET',
   # 'R_CURLY_BRACKET',
)


def get_lexer(module=None):
    t_IEML_OBJECT = r'.+'
    # t_L_CURLY_BRACKET = r'\{'
    # t_R_CURLY_BRACKET = r'\}'

    t_ignore  = ' \t\n'

    # Error handling rule
    def t_error(t):
        print("Illegal character '%s'" % t.value[0])
        t.lexer.skip(1)

    return lxr.lex(module=module, errorlog=logging)


if __name__ == "__main__":
    # Test it out
    data = '{/[([a.i.-] + [i.i.-]) * ([E:A:T:.]+[E:S:.wa.-]+[E:S:.o.-])]/' \
           '/[([a.i.-] + [i.i.-]) * ([E:A:T:.]+[E:S:.wa.-]+[E:S:.o.-])]/<sup dude>}'

    # Give the lexer some input
    lexer = get_lexer()
    lexer.input(data)

    # Tokenize
    while True:
        tok = lexer.token()
        if not tok:
            break      # No more input
        print(tok)




from .exceptions import InvalidIEMLObjectArgument


class cached_property:
    def __init__(self, factory):
        self._factory = factory
        self._attr_name = factory.__name__

    def __get__(self, instance, owner):
        attr = self._factory(instance)
        setattr(instance, self._attr_name, attr)
        return attr

class TreeStructure:
    def __init__(self):
        super().__init__()
        self._str = None
        self._paths = None
        self.children = None  # will be an iterable (list or tuple)

    def __str__(self):
        return self._str

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if not isinstance(other, (TreeStructure, str)):
            return False

        return self._str == str(other)

    def __hash__(self):
        """Since the IEML string for any proposition AST is supposed to be unique, it can be used as a hash"""
        return self.__str__().__hash__()

    def __iter__(self):
        """Enables the syntactic sugar of iterating directly on an element without accessing "children" """
        return self.children.__iter__()

    def tree_iter(self):
        yield self
        for c in self.children:
            yield from c.tree_iter()


class IEMLType(type):
    """This metaclass enables the comparison of class times, such as (Sentence > Word) == True"""

    def __init__(cls, name, bases, dct):
        child_list = ['Term', 'Morpheme', 'Word', 'Clause', 'Sentence',
                     'SuperClause', 'SuperSentence', 'Text', 'Hyperlink', 'Hypertext']
        if name in child_list:
            cls.__rank = child_list.index(name) + 1
        else:
            cls.__rank = 0

        super(IEMLType, cls).__init__(name, bases, dct)

    def __hash__(self):
        return self.__rank

    def __eq__(self, other):
        if isinstance(other, IEMLType):
            return self.__rank == other.__rank
        else:
            return False

    def __ne__(self, other):
        return not IEMLType.__eq__(self, other)

    def __gt__(self, other):
        return IEMLType.__ge__(self, other) and self != other

    def __le__(self, other):
        return IEMLType.__lt__(self, other) and self == other

    def __ge__(self, other):
        return not IEMLType.__lt__(self, other)

    def __lt__(self, other):
        return self.__rank < other.__rank

    def ieml_rank(self):
        return self.__rank


class IEMLObjects(TreeStructure, metaclass=IEMLType):
    closable = False

    def __init__(self, children, literals=None):
        super().__init__()
        self.children = tuple(children)

        _literals = []
        if literals is not None:
            if isinstance(literals, str):
                _literals = [literals]
            else:
                try:
                    _literals = tuple(literals)
                except TypeError:
                    raise InvalidIEMLObjectArgument(self.__class__, "The literals argument %s must be an iterable of "
                                                                    "str or a str."%str(literals))

        self.literals = tuple(_literals)
        self._do_precompute_str()

    def __gt__(self, other):
        if not isinstance(other, IEMLObjects):
            raise NotImplemented

        if self.__class__ != other.__class__:
            return self.__class__ > other.__class__

        return self._do_gt(other)

    def _do_gt(self, other):
        return self.children > other.children

    def compute_str(self, children_str):
        return '#'.join(children_str)

    def _compute_str(self):
        if self._str is not None:
            return self._str
        _literals = ''
        if self.literals:
            _literals = '<' + '><'.join(self.literals) + '>'
        return self.compute_str([e._compute_str() for e in self.children]) + _literals

    def _do_precompute_str(self):
        self._str = self._compute_str()



import os
import pickle
from enum import Enum, unique, IntEnum
from itertools import combinations, product, groupby, chain

import bidict
import numpy as np
from scipy.sparse.csr import csr_matrix

from ieml.dictionary.dictionary import Dictionary

def default_metric(dictionary_version):
    mat = get_matrix('distance', dictionary_version)
    return lambda t0, t1: mat[t0.index, t1.index]


def distance_pack(t0, version):
    reltypes = get_matrix('relation', version)
    distance = get_matrix('distance', version)

    res = [(reltypes[t0.index, i], t0.dictionary.index[i]) for i in reltypes[t0.index, :].indices]
    res = sorted(res, key=lambda t: t[0])

    return [(RelationType(key).name, [t[1] for t in v]) for key, v in groupby(res, key=lambda t: t[0])]


def rank_from_term(term, metric, nb_terms=30):
    res = sorted([(metric(term, t), t) for t in term.dictionary], reverse=True)[:nb_terms]
    return [r for r in res if r[0] if r[0] != 1.0]


def test_metric(metric, t0, n=30):
    def _str_term(t):
        return "%s - (%s)"%(str(t), t.translations['fr'])

    reltypes = get_matrix('relation', t0.dictionary.version)

    print("distance from %s"%_str_term(t0))
    res = rank_from_term(t0, metric, nb_terms=n)

    print('\n'.join("[%.3f] [%s]: %s"%(r[0], RelationType(reltypes[t0.index, r[1].index]).name,_str_term(r[1])) for r in res))



MATRIX_CACHE = {}


def get_matrix(name, version):
    file = '/tmp/cache_%s_%s.npy' % (name, str(version))
    if file in MATRIX_CACHE:
        return MATRIX_CACHE[file]
    elif os.path.isfile(file):
        with open(file, 'rb') as fp:
            return pickle.load(fp)
    else:
        print("\t[*] Building distance matrix %s."%name)
        mat = MATRIX_BUILD[name](version)
        for k, v in mat.items():
            name = '/tmp/cache_%s_%s.npy' % (k, str(version))
            MATRIX_CACHE[name] = v
            with open(name, 'wb') as fp:
                pickle.dump(v, fp)

        return MATRIX_CACHE[file]
# to set

#
# |____|____|____|____|______|_Ranks______________________|________|
# |Ass  Opp  Crss Twin|Father|         |GFather|          |GGFather|
# |                   |Child |         |GChild |          |GGChild |
# 0                   C     C+k        (1+C+-k)/2        1-k       1

def _relation_value(reltype, max_rank=None):
    C = 1. / 3
    k = 1. / 8

    def _sam_to_float(reltype):
        MAP = {'s': 1, 'a': 2, 'm': 3}
        res = 1.
        for c in reltype:
            res *= MAP[c] / 3.
        return res

    if reltype == 'Nll':
        return 0, RELATIONS_TYPES, RelationType.Null
    elif reltype == 'associated':
        return C * 1. / 4, RELATIONS_TYPES[reltype], RelationType.Associated
    elif reltype == 'opposed':
        return C * 1. / 2, RELATIONS_TYPES[reltype], RelationType.Opposed
    elif reltype == 'crossed':
        return C * 3. / 4, RELATIONS_TYPES[reltype], RelationType.Crossed
    elif reltype == 'twin':
        return C, RELATIONS_TYPES[reltype], RelationType.Twin
    elif reltype.startswith('table_'):
        i = int(reltype[6:7])
        # rank
        # available slot: 2, 4
        ratio = float(max_rank - i) / max_rank
        slot = 2 if ratio < 0.5 else 4
        return C + k + (1 - 2*k - C) * ratio, (slot, max_rank - i), RelationType['Rank_%d'%i]
    else:
        #sam
        if len(reltype) == 1:
            return C + k * _sam_to_float(reltype), RELATIONS_TYPES[reltype], RelationType['Etymology_%s'%reltype]
        elif len(reltype) == 2:
            return (1 + C - k) / 2 + k * _sam_to_float(reltype), RELATIONS_TYPES[reltype], RelationType['Etymology_%s'%reltype]
        elif len(reltype) == 3:
            return 1 - k + k * _sam_to_float(reltype), RELATIONS_TYPES[reltype], RelationType['Etymology_%s'%reltype]
        else:
            ValueError("Invalid relation %s"%str(reltype))


def _build_distance_matrix(version):
    def _enumerate_ancestors(t, prefix=''):
        for k, v in t.relations.father.items():
            for t1 in v:
                # if t1 is layer 0, we include this etymology only if it is a direct father/child
                if t1.layer == 0 and len(prefix) != 0:
                    continue

                yield (prefix + k, t1)

                if len(prefix) < 2:
                    yield from _enumerate_ancestors(t1, prefix=prefix + k)


    d = Dictionary(version)
    # relation_matrix = np.empty(shape=(len(d),) * 2,
    #                         dtype=[('distance', 'f4'),('slot', 'i4', (2,)), ('reltype', 'S16')])

    distance_matrix = np.empty(shape=(len(d),) * 2, dtype='f4')
    order_matrix = np.empty(shape=(len(d),) * 2, dtype='i4')
    relation_type_matrix = np.empty(shape=(len(d),) * 2, dtype='i4')

    # relation_matrix['reltype'] = 'none'
    # relation_matrix['slot'] = (0, 0)
    # relation_matrix['distance'] = 1.0

    for root in d.roots:
        max_rank = max(t.rank for t in root.relations.contains if len(t) != 1) + 1

        for t0, t1 in combinations(root.relations.contains, 2):

            reltype = min({r for r in t0.relations.to(t1) if r not in ('contained', 'contains')},
                          key=lambda r: RELATIONS_TYPES[r])


            v, order, reltype = _relation_value(reltype, max_rank=max_rank)
            distance_matrix[t0.index, t1.index] = 1 - v
            order_matrix[t0.index, t1.index] = order[0] * 100 + order[1]
            relation_type_matrix[t0.index, t1.index] = int(reltype)

    for layer in d.layers:
        for t0 in layer:
            for prefix, t1 in _enumerate_ancestors(t0):
                v, order, reltype = _relation_value(prefix)
                distance_matrix[t0.index, t1.index] = 1 - v
                order_matrix[t0.index, t1.index] = order[0] * 100 + order[1]
                relation_type_matrix[t0.index, t1.index] = int(reltype)

    distance_matrix = distance_matrix + distance_matrix.transpose()
    order_matrix = order_matrix + order_matrix.transpose()
    relation_type_matrix = relation_type_matrix + relation_type_matrix.transpose()

    for i in range(len(d)):
        distance_matrix[i, i] = 1.0
        order_matrix[i, i] = 0
        relation_type_matrix[i, i] = int(RelationType.Equal)

    distance_matrix = csr_matrix(distance_matrix)
    order_matrix = csr_matrix(order_matrix)
    relation_type_matrix = csr_matrix(relation_type_matrix)

    return {'distance' : distance_matrix,
            'order': order_matrix,
            'relation': relation_type_matrix}





# Ordre sur les relations:
# Class principals :
# Ass > Opp > Crs > Twn > Fth > Rnk > Chd > Nll
# Sous-class
# Fth/Chd :
# S > A > M > SS > SA ...
# Rnk:
# 5 > 4 ... > 0


def _enumerate_ancestors(t, prefix=''):
    for k, v in t.relations.father.items():
        for t1 in v:
            # if t1 is layer 0, we include this etymology only if it is a direct father/child
            if t1.layer == 0 and len(prefix) != 0:
                continue

            yield (prefix + k, t1)
            yield from _enumerate_ancestors(t1, prefix=prefix + k)


def _build_ancestor_matrix(version):
    d = Dictionary(version)
    matrix = np.empty(shape=(len(d),) * 2, dtype=[('relation', "U8"), ('prefix', "U8")])

    for layer in d.layers:
        for t0 in layer:
            for prefix, t1 in _enumerate_ancestors(t0):
                matrix[t0.index, t1.index]['relation'] = 'Fth'
                matrix[t0.index, t1.index]['prefix'] = prefix

                matrix[t1.index, t0.index]['relation'] = 'Chd'
                matrix[t1.index, t0.index]['prefix'] = prefix

    return {'ancestor': matrix}


MATRIX_BUILD = {
    'distance': _build_distance_matrix,
    'order': _build_distance_matrix,
    'relation': _build_distance_matrix,
    'ancestor': _build_ancestor_matrix
}



# index to order
# 1st main
# 0 <= 2nd < 100
RELATIONS_TYPES = {
    'associated': (0, 1),
    'opposed': (0, 2),
    'crossed': (0, 3),
    'twin': (0, 4),
    'table_5': (4, 0),
    'table_4': (4, 1),
    'table_3': (4, 2),
    'table_2': (4, 3),
    'table_1': (4, 4),
    'table_0': (4, 5),
    **{''.join(s): (1 + 2*(i-1), j) for i in range(1, 4) for j, s in enumerate(product('sam', repeat=i))}
}


RelationType = unique(IntEnum('RelationType', {
    'Null': 0,
    'Equal': 1,
    'Crossed': 2,
    'Associated': 3,
    'Twin': 4,
    'Opposed': 5,
    **{'Rank_%d'%i: 6 + i  for i in range(6)},
    **{'Etymology_%s'%''.join(s): int(11 + (3 ** i - 1) / 2 + j) for i in range(1, 4) for j, s in enumerate(product('sam', repeat=i))}
}))




# class RelationTerm:
#     ancestor = None
#     def __init__(self, term0, term1):
#         if self.ancestor is None:
#             self.__class__.ancestor = get_matrix('ancestor', term0.dictionary.version)
#
#         self.t0 = term0
#         self.t1 = term1
#         if self.t0 == self.t1:
#             self.relation_type = 'Idt'
#             self.order = 0
#
#         elif self.t0.root == self.t1.root:
#             self.relation_type = min({r for r in self.t0.relations.to(self.t1) if r not in ('contained', 'contains')},
#                           key=lambda r: RELATIONS_TYPES[r])
#
#             self.order = RELATIONS_TYPES[self.relation_type]
#         elif self.ancestor[self.t0.index, self.t1.index]['relation'] != '' and\
#              len(self.ancestor[self.t0.index, self.t1.index]['prefix']) < 4:
#             self.relation_type, self.subtype = self.ancestor[self.t0.index, self.t1.index]
#
#             self.order = RELATIONS_TYPES[self.subtype]
#         else:
#             self.relation_type = 'Nll'
#             _, self.order = _relation_value('Nll')
#
#     def __eq__(self, other):
#         return {self.t0, self.t1} == {other.t0, other.t1}
#
#     def __lt__(self, other):
#         return self.order < other.order
#
#     def __str__(self):
#         return _RELATION_TYPES[self.relation_type]['name']
#



from configparser import ExtendedInterpolation
import os
from os.path import isfile, isdir, expanduser, join, dirname
import configparser

_config = configparser.ConfigParser(interpolation=ExtendedInterpolation())
_config.read(join(dirname(__file__), 'default_config.conf'))

_HOME_FOLDER = expanduser("~")
ieml_folder = join(_HOME_FOLDER, _config.get('DEFAULT', 'iemlfolder'))
_config_file = join(ieml_folder, _config.get('DEFAULT', 'configfile'))
parser_folder = join(ieml_folder, 'parser')


if not isdir(ieml_folder):
    os.mkdir(ieml_folder)

if not isdir(parser_folder):
    os.mkdir(parser_folder)

if isfile(_config_file):
    _config.read(_config_file)


def get_configuration():
    return _config

from . import tools
from . import exceptions
from . import constants
from . import syntax
from . import dictionary
from . import usl




import logging
import os
import ply.yacc as yacc

from ieml.dictionary.terms import Term
from ieml.dictionary.tools import term
from ... import parser_folder
from ...metaclasses import Singleton
from ...exceptions import CannotParse
from .. import Word, Morpheme, Clause, SuperClause, Sentence, SuperSentence, Text, Hypertext, Hyperlink, PropositionPath
from ...dictionary import Dictionary

from .lexer import get_lexer, tokens


def _add(lp1, p2):
    return lp1[0] + [p2[0]], lp1[1] + p2[1]


def _build(*arg):
    if len(arg) > 1:
        _h = []
        for e in arg[1:]:
            for h in e[1]:
                h[1].insert(0, arg[0])
                _h.append(h)
        return arg[0], _h
    return arg[0], []


def _hyperlink(node, text_list):
    return node[0], node[1] + [(text, [node[0]]) for text in text_list[0]]


class IEMLParser(metaclass=Singleton):
    tokens = tokens

    def __init__(self):

        # Build the lexer and parser
        self.lexer = get_lexer()
        self.parser = yacc.yacc(module=self, errorlog=logging, start='proposition',
                                debug=False, optimize=True, picklefile=os.path.join(parser_folder, "ieml_parser.pickle"))
        self._ieml = None

    def parse(self, s):
        """Parses the input string, and returns a reference to the created AST's root"""
        self._ieml = s
        self.root = None
        self.hyperlinks = []
        self.parser.parse(s, lexer=self.lexer)

        if self.root is not None:
            if self.hyperlinks:
                self.root = Hypertext(self.hyperlinks)

            # if isinstance(self.root, Term):
            #     self.root = Word(root=Morpheme([self.root]))

            return self.root
        else:
            raise CannotParse(s)

    # Parsing rules
    def p_ieml_proposition(self, p):
        """proposition :  script
                        | term
                        | morpheme
                        | word
                        | clause
                        | sentence
                        | superclause
                        | supersentence
                        | text"""
        self.root = p[1][0]

    def p_literal_list(self, p):
        """literal_list : literal_list LITERAL
                        | LITERAL"""

        if len(p) == 3:
            p[0] = p[1] + [p[2][1:-1]]
        else:
            p[0] = [p[1][1:-1]]


    def p_script(self, p):
        """script : TERM """
        p[0] = _build(term(p[1]))


    def p_term(self, p):
        """term : LBRACKET script RBRACKET"""
        p[0] = p[2]

    def p_proposition_sum(self, p):
        """terms_sum : terms_sum PLUS term
                    | term
            clauses_sum : clauses_sum PLUS clause
                    | clause
            superclauses_sum : superclauses_sum PLUS superclause
                    | superclause
            closed_proposition_list : closed_proposition_list closed_proposition
                                    | closed_proposition"""
        if len(p) == 4:
            p[0] = _add(p[1], p[3])
        elif len(p) == 3:
            p[0] = _add(p[1], p[2])
        else:
            p[0] = _add(([], []), p[1])

    def p_morpheme(self, p):
        """morpheme : LPAREN terms_sum RPAREN"""
        p[0] = _build(Morpheme(p[2][0]), p[2])

    def p_word(self, p):
        """word : LBRACKET morpheme RBRACKET
                | LBRACKET morpheme RBRACKET literal_list
                | LBRACKET morpheme TIMES morpheme RBRACKET
                | LBRACKET morpheme TIMES morpheme RBRACKET literal_list"""

        if len(p) == 4:
            p[0] = _build(Word(root=p[2][0]), p[2])
        elif len(p) == 5:
            p[0] = _build(Word(root=p[2][0], literals=p[4]), p[2])
        elif len(p) == 6:
            p[0] = _build(Word(root=p[2][0], flexing=p[4][0]), p[2], p[4])
        else:
            p[0] = _build(Word(root=p[2][0], flexing=p[4][0], literals=p[6]), p[2], p[4])

    def p_decorated(self, p):
        """decorated_word : word
                           | word text_list
            decorated_sentence : sentence
                                | sentence text_list
            decorated_supersentence : supersentence
                                    | supersentence text_list"""
        if len(p) == 2:
            p[0] = p[1]
        else:
            p[0] = _hyperlink(p[1], p[2])

    def p_clause(self, p):
        """clause : LPAREN decorated_word TIMES decorated_word TIMES decorated_word RPAREN"""
        p[0] = _build(Clause(substance=p[2][0], attribute=p[4][0], mode=p[6][0]), p[2], p[4], p[6])

    def p_sentence(self, p):
        """sentence : LBRACKET clauses_sum RBRACKET
                    | LBRACKET clauses_sum RBRACKET literal_list"""
        if len(p) == 4:
            p[0] = _build(Sentence(p[2][0]), p[2])
        else:
            p[0] = _build(Sentence(p[2][0], literals=p[4]), p[2])

    def p_superclause(self, p):
        """superclause : LPAREN decorated_sentence TIMES decorated_sentence TIMES decorated_sentence RPAREN"""
        p[0] = _build(SuperClause(substance=p[2][0], attribute=p[4][0], mode=p[6][0]), p[2], p[4], p[6])

    def p_super_sentence(self, p):
        """supersentence : LBRACKET superclauses_sum RBRACKET
                         | LBRACKET superclauses_sum RBRACKET literal_list"""
        if len(p) == 4:
            p[0] = _build(SuperSentence(p[2][0]), p[2])
        else:
            p[0] = _build(SuperSentence(p[2][0], literals=p[4]), p[2])

    def p_closed_proposition(self, p):
        """ closed_proposition : SLASH decorated_word SLASH
                               | SLASH decorated_sentence SLASH
                               | SLASH decorated_supersentence SLASH"""
        p[0] = p[2]

    def p_text(self, p):
        """text : L_CURLY_BRACKET closed_proposition_list R_CURLY_BRACKET"""
        p[0] = _build(Text(p[2][0]))

        if p[2][1]:
            # tuple of the hyperlink (end text, path)
            self.hyperlinks += [Hyperlink(p[0][0], e[0], PropositionPath(e[1])) for e in p[2][1]]

    def p_text_list(self, p):
        """text_list : text_list text
                    | text"""
        if len(p) == 3:
            p[0] = _add(p[1], p[2])
        else:
            p[0] = _add(([], []), p[1])

    def p_error(self, p):
        if p:
            print("Syntax error at '%s' (%d, %d)" % (p.value, p.lineno, p.lexpos))
        else:
            print("Syntax error at EOF")

        raise CannotParse(self._ieml)

if __name__ == '__main__':
    print(str(IEMLParser().parse('[wa.]')))
